<?php

namespace Drupal\auditfiles\Services;

use Drupal\auditfiles\Batch\AuditFilesBatchProcess;
use Drupal\auditfiles\Batch\AuditFilesNotOnServerBatchProcess;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Messenger\MessengerTrait;

/**
 * Providing the service that used in not in database functionality.
 */
final class AuditFilesNotOnServer {

  use MessengerTrait;

  /**
   * Constructs a new AuditFilesNotOnServer.
   */
  final public function __construct(
    MessengerInterface $messenger,
    protected ConfigFactoryInterface $configFactory,
    protected Connection $connection,
    protected DateFormatterInterface $dateFormatter,
    protected FileSystemInterface $fileSystem,
  ) {
    $this->setMessenger($messenger);
  }

  /**
   * Retrieves the file IDs to operate on.
   *
   * @return array
   *   The file IDs.
   */
  public function auditfilesNotOnServerGetFileList(): array {
    $config = $this->configFactory->get('auditfiles.settings');
    $file_ids = [];
    $maximum_records = $config->get('auditfiles_report_options_maximum_records');
    $query = $this->connection->select('file_managed', 'fm');
    $query->orderBy('changed', 'DESC');
    $query->range(0, $maximum_records);
    $query->fields('fm', ['fid', 'uri']);
    $results = $query->execute()->fetchAll();
    foreach ($results as $result) {
      $target = $this->fileSystem->realpath($result->uri);
      if (!file_exists($target)) {
        $file_ids[] = $result->fid;
      }
    }
    return $file_ids;
  }

  /**
   * Retrieves information about an individual file from the database.
   *
   * @param int $file_id
   *   The ID of the file to prepare for display.
   * @param string $date_format
   *   The Date format to prepair for display.
   *
   * @return array
   *   The row for the table on the report, with the file's
   *   information formatted for display.
   */
  public function auditfilesNotOnServerGetFileData($file_id, $date_format): array {
    $query = $this->connection->select('file_managed', 'fm');
    $query->condition('fm.fid', $file_id);
    $query->fields('fm', [
      'fid',
      'uid',
      'filename',
      'uri',
      'filemime',
      'filesize',
      'created',
      'status',
    ]);
    $results = $query->execute()->fetchAll();
    $file = $results[0];
    return [
      'fid' => $file->fid,
      'uid' => $file->uid,
      'filename' => $file->filename,
      'uri' => $file->uri,
      'path' => $this->fileSystem->realpath($file->uri),
      'filemime' => $file->filemime,
      'filesize' => $file->filesize ? number_format($file->filesize) : NULL,
      'datetime' => $this->dateFormatter->format($file->created, $date_format),
      'status' => ($file->status == 1) ? 'Permanent' : 'Temporary',
    ];
  }

  /**
   * Returns the header to use for the display table.
   *
   * @return array
   *   The header to use.
   */
  public function auditfilesNotOnServerGetHeader(): array {
    return [
      'fid' => [
        'data' => t('File ID'),
      ],
      'uid' => [
        'data' => t('User ID'),
      ],
      'filename' => [
        'data' => t('Name'),
      ],
      'uri' => [
        'data' => t('URI'),
      ],
      'path' => [
        'data' => t('Path'),
      ],
      'filemime' => [
        'data' => t('MIME'),
      ],
      'filesize' => [
        'data' => t('Size'),
      ],
      'datetime' => [
        'data' => t('When added'),
      ],
      'status' => [
        'data' => t('Status'),
      ],
    ];
  }

  /**
   * Creates the batch for deleting files from the database.
   *
   * @param array $fileids
   *   The list of file IDs to be processed.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The definition of the batch.
   */
  public function auditfilesNotOnServerBatchDeleteCreateBatch(array $fileids): BatchBuilder {
    $batch = (new BatchBuilder())
      ->setTitle(\t('Deleting files from the database'))
      ->setErrorMessage(\t('One or more errors were encountered processing the files.'))
      ->setFinishCallback([AuditFilesBatchProcess::class, 'finishBatch'])
      ->setProgressMessage(\t('Completed @current of @total operations.'));
    foreach ($fileids as $file_id) {
      if ($file_id != 0) {
        $batch->addOperation([AuditFilesNotOnServerBatchProcess::class, 'auditfilesNotOnServerBatchDeleteProcessBatch'], [$file_id]);
      }
    }
    return $batch;
  }

  /**
   * Deletes the specified file from the database.
   *
   * @param int $file_id
   *   The ID of the file to delete from the database.
   */
  public function auditfilesNotOnServerBatchDeleteProcessFile($file_id): void {
    $num_rows = $this->connection->delete('file_managed')
      ->condition('fid', $file_id)
      ->execute();
    if (empty($num_rows)) {

      $this->messenger()->addWarning(
        t(
          'There was a problem deleting the record with file ID %fid from the file_managed table. Check the logs for more information.',
          ['%fid' => $file_id]
        )
      );
    }
    else {
      $this->messenger()->addStatus(
        t(
          'Sucessfully deleted File ID : %fid from the file_managed table.',
          ['%fid' => $file_id]
        )
      );
    }
  }

}
