<?php

namespace Drupal\auditfiles\Services;

use Drupal\auditfiles\Batch\AuditFilesBatchProcess;
use Drupal\auditfiles\Batch\AuditFilesReferencedNotUsedBatchProcess;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Url;

/**
 * List all methods used in referenced not used functionality.
 */
final class AuditFilesReferencedNotUsed {

  use MessengerTrait;

  /**
   * Constructs a new AuditFilesReferencedNotUsed.
   */
  final public function __construct(
    MessengerInterface $messenger,
    protected ConfigFactoryInterface $configFactory,
    protected Connection $connection,
    protected EntityFieldManagerInterface $entityFieldManager,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected FileUrlGeneratorInterface $fileUrlGenerator,
  ) {
    $this->setMessenger($messenger);
  }

  /**
   * Retrieves the file IDs to operate on.
   *
   * @return array
   *   The file IDs.
   */
  public function auditfilesReferencedNotUsedGetFileList(): array {
    $config = $this->configFactory->get('auditfiles.settings');
    $files_referenced = [];
    // Get a list of all files that are referenced in content.
    $fields = $field_data = [];
    $fields[] = $this->entityFieldManager->getFieldMapByFieldType('image');
    $fields[] = $this->entityFieldManager->getFieldMapByFieldType('file');
    if ($fields) {
      $count = 0;
      foreach ($fields as $value) {
        foreach ($value as $table_prefix => $entity_type) {
          foreach ($entity_type as $key1 => $value1) {
            $field_data[$count]['table'] = $table_prefix . '__' . $key1;
            $field_data[$count]['column'] = $key1 . '_target_id';
            $field_data[$count]['entity_type'] = $table_prefix;
            $count++;
          }
        }
      }
      foreach ($field_data as $value) {
        $table = $value['table'];
        $column = $value['column'];
        $entity_type = $value['entity_type'];
        if ($this->connection->schema()->tableExists($table)) {
          $fu_query = $this->connection->select('file_usage', 'fu')->fields('fu', ['fid'])->execute()->fetchCol();
          $query = $this->connection->select($table, 't')->fields('t', ['entity_id', $column]);
          if (!empty($fu_query)) {
            $query->condition('t.' . $column, $fu_query, 'NOT IN');
          }
          $maximum_records = $config->get('auditfiles_report_options_maximum_records');
          if ($maximum_records > 0) {
            $query->range(0, $maximum_records);
          }
          $file_references = $query->execute()->fetchAll();
          foreach ($file_references as $file_reference) {
            $reference_id = $table . '.' . $column . '.' . $file_reference->entity_id . '.' . $entity_type . '.' . $file_reference->{$column};
            $files_referenced[$reference_id] = [
              'table' => $table,
              'column' => $column,
              'entity_id' => $file_reference->entity_id,
              'file_id' => $file_reference->{$column},
              'entity_type' => $entity_type,
            ];
          }
        }
      }
    }
    return $files_referenced;
  }

  /**
   * Retrieves information about an individual file from the database.
   *
   * @param array $row_data
   *   The data to use for creating the row.
   *
   * @return array
   *   The row for the table on the report, with the file's
   *   information formatted for display.
   */
  public function auditfilesReferencedNotUsedGetFileData(array $row_data): array {
    $query = 'SELECT * FROM {' . $row_data['table'] . '} WHERE ' . $row_data['column'] . ' = ' . $row_data['file_id'];
    $result = $this->connection->query($query)->fetchAll();
    $result = reset($result);
    if ($row_data['entity_type'] == 'node') {
      $url = Url::fromUri('internal:/node/' . $result->entity_id);
      $entity_id_display = Link::fromTextAndUrl('node/' . $result->entity_id, $url)->toString();
    }
    else {
      $entity_id_display = $result->entity_id;
    }
    $row = [
      'file_id' => $result->{$row_data['column']},
      'entity_type' => $row_data['entity_type'],
      'bundle' => ['data' => $result->bundle, 'hidden' => TRUE],
      'entity_id' => ['data' => $result->entity_id, 'hidden' => TRUE],
      'entity_id_display' => $entity_id_display,
      'field' => $row_data['table'] . '.' . $row_data['column'],
      'table' => ['data' => $row_data['table'], 'hidden' => TRUE],
      'uri' => 'No file object exists for this reference.',
      'filename' => ['data' => '', 'hidden' => TRUE],
      'filemime' => '--',
      'filesize' => '--',
    ];
    // If there is a file in the file_managed table, add some of that
    // information to the row, too.
    $file_managed = $this->entityTypeManager->getStorage('file')->load($result->{$row_data['column']});
    if (!empty($file_managed)) {
      $row['uri'] = Link::fromTextAndUrl($file_managed->getFileuri(), Url::fromUri($this->fileUrlGenerator->generateAbsoluteString($file_managed->getFileuri()), ['attributes' => ['target' => '_blank']]));
      $row['filename'] = ['data' => $file_managed->getFilename(), 'hidden' => TRUE];
      $row['filemime'] = $file_managed->getMimeType();
      $row['filesize'] = $file_managed->getSize();
    }
    return $row;
  }

  /**
   * Returns the header to use for the display table.
   *
   * @return array
   *   The header to use.
   */
  public function auditfilesReferencedNotUsedGetHeader(): array {
    return [
      'file_id' => [
        'data' => t('File ID'),
      ],
      'entity_type' => [
        'data' => t('Referencing entity type'),
      ],
      'entity_id_display' => [
        'data' => t('Referencing entity ID'),
      ],
      'field' => [
        'data' => t('Field referenced in'),
      ],
      'uri' => [
        'data' => t('URI'),
      ],
      'filemime' => [
        'data' => t('MIME'),
      ],
      'filesize' => [
        'data' => t('Size (in bytes)'),
      ],
    ];
  }

  /**
   * Creates the batch for adding files to the file_usage table.
   *
   * @param array $referenceids
   *   The list of IDs to be processed.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The definition of the batch.
   */
  public function auditfilesReferencedNotUsedBatchAddCreateBatch(array $referenceids): BatchBuilder {
    $batch = (new BatchBuilder())
      ->setTitle(\t('Adding files to the file_usage table'))
      ->setErrorMessage(\t('One or more errors were encountered processing the files.'))
      ->setFinishCallback([AuditFilesBatchProcess::class, 'finishBatch'])
      ->setProgressMessage(\t('Completed @current of @total operations.'));
    foreach ($referenceids as $reference_id) {
      if (!empty($reference_id)) {
        $batch->addOperation([AuditFilesReferencedNotUsedBatchProcess::class, 'auditfilesReferencedNotUsedBatchAddProcessBatch'], [$reference_id]);
      }
    }
    return $batch;
  }

  /**
   * Adds the specified file to the file_usage table.
   *
   * @param string $reference_id
   *   The ID for keeping track of the reference.
   */
  public function auditfilesReferencedNotUsedBatchAddProcessFile($reference_id): void {
    $reference_id_parts = explode('.', $reference_id);
    $data = [
      'fid' => $reference_id_parts[4],
      // @todo This is hard coded for now, but need to determine how to figure out
      // which module needs to be here.
      'module' => 'file',
      'type' => $reference_id_parts[3],
      'id' => $reference_id_parts[2],
      'count' => 1,
    ];
    // Make sure the file is not already in the database.
    $query = 'SELECT fid FROM {file_usage}
    WHERE fid = :fid AND module = :module AND type = :type AND id = :id';
    $existing_file = $this->connection->query(
      $query,
      [
        ':fid' => $data['fid'],
        ':module' => $data['module'],
        ':type' => $data['type'],
        ':id' => $data['id'],
      ]
    )->fetchAll();
    if (empty($existing_file)) {
      // The file is not already in the database, so add it.
      $this->connection->insert('file_usage')->fields($data)->execute();
    }
    else {
      $this->messenger->addError(
        t(
           'The file is already in the file_usage table (file id: "@fid", module: "@module", type: "@type", entity id: "@id").',
          [
            '@fid' => $data['fid'],
            '@module' => $data['module'],
            '@type' => $data['type'],
            '@id' => $data['id'],
          ]
        )
      );
    }
  }

  /**
   * Creates the batch for deleting file references from their content.
   *
   * @param array $referenceids
   *   The list of IDs to be processed.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The definition of the batch.
   */
  public function auditfilesReferencedNotUsedBatchDeleteCreateBatch(array $referenceids): BatchBuilder {
    $batch = (new BatchBuilder())
      ->setTitle(\t('Deleting file references from their content'))
      ->setErrorMessage(\t('One or more errors were encountered processing the files.'))
      ->setFinishCallback([AuditFilesBatchProcess::class, 'finishBatch'])
      ->setProgressMessage(\t('Completed @current of @total operations.'));
    foreach ($referenceids as $reference_id) {
      if ($reference_id != '') {
        $batch->addOperation([AuditFilesReferencedNotUsedBatchProcess::class, 'auditfilesReferencedNotUsedBatchDeleteProcessBatch'], [$reference_id]);
      }
    }
    return $batch;
  }

  /**
   * Deletes the specified file from the database.
   *
   * @param string $reference_id
   *   The ID for keeping track of the reference.
   */
  public function auditfilesReferencedNotUsedBatchDeleteProcessFile($reference_id): void {
    $reference_id_parts = explode('.', $reference_id);
    $num_rows = $this->connection->delete($reference_id_parts[0])
      ->condition($reference_id_parts[1], $reference_id_parts[4])
      ->execute();
    if (empty($num_rows)) {
      $this->messenger()->addWarning(
        t(
          'There was a problem deleting the reference to file ID %fid in the %entity_type with ID %eid. Check the logs for more information.',
          [
            '%fid' => $reference_id_parts[4],
            '%entity_type' => $reference_id_parts[3],
            '%eid' => $reference_id_parts[2],
          ]
        )
      );
    }
    else {
      $this->messenger()->addStatus(
        t(
          'file ID %fid  deleted successfully.',
          [
            '%fid' => $reference_id_parts[4],
          ]
        )
      );
    }
  }

}
