<?php

namespace Drupal\auditfiles\Services;

use Drupal\auditfiles\Batch\AuditFilesBatchProcess;
use Drupal\auditfiles\Batch\AuditFilesUsedNotManagedBatchProcess;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Url;

/**
 * Form for Files used not managed functionality.
 */
final class AuditFilesUsedNotManaged {

  use MessengerTrait;

  /**
   * Constructs a new AuditFilesUsedNotManaged.
   */
  final public function __construct(
    MessengerInterface $messenger,
    protected ConfigFactoryInterface $configFactory,
    protected Connection $connection,
  ) {
    $this->setMessenger($messenger);
  }

  /**
   * Retrieves the file IDs to operate on.
   *
   * @return array
   *   The file IDs.
   */
  public function auditfilesUsedNotManagedGetFileList() {
    // Get all the file IDs in the file_usage table that are not in the
    // file_managed table.
    $config = $this->configFactory->get('auditfiles.settings');
    $fm_query = $this->connection->select('file_managed', 'fm')->fields('fm', ['fid'])->execute()->fetchCol();
    $query = $this->connection->select('file_usage', 'fu')->fields('fu', ['fid']);
    if (!empty($fm_query)) {
      $query->condition('fu.fid', $fm_query, 'NOT IN');
    }
    $maximum_records = $config->get('auditfiles_report_options_maximum_records');
    if ($maximum_records > 0) {
      $query->range(0, $maximum_records);
    }
    return $query->execute()->fetchCol();
  }

  /**
   * Retrieves information about an individual file from the database.
   */
  public function auditfilesUsedNotManagedGetFileData($file_id): array {
    // Get the file information for the specified file ID from the database.
    $query = 'SELECT * FROM {file_usage} WHERE fid = :file_id';
    $file = $this->connection->query($query, ['file_id' => $file_id])->fetchObject();

    $url = Url::fromUri('internal:/' . $file->type . '/' . $file->id);
    $result = Link::fromTextAndUrl($file->type . '/' . $file->id, $url)->toString();
    return [
      'fid' => $file->fid,
      'module' => $file->module . ' ' . t('module'),
      'id' => $result,
      'count' => $file->count,
    ];
  }

  /**
   * Returns the header to use for the display table.
   */
  public function auditfilesUsedNotManagedGetHeader(): array {
    return [
      'fid' => [
        'data' => t('File ID'),
      ],
      'module' => [
        'data' => t('Used by'),
      ],
      'id' => [
        'data' => t('Used in'),
      ],
      'count' => [
        'data' => t('Count'),
      ],
    ];
  }

  /**
   * Creates the batch for deleting files from the file_usage table.
   */
  public function auditfilesUsedNotManagedBatchDeleteCreateBatch(array $fileids): BatchBuilder {
    $batch = (new BatchBuilder())
      ->setTitle(\t('Deleting files from the file_usage table'))
      ->setErrorMessage(\t('One or more errors were encountered processing the files.'))
      ->setFinishCallback([AuditFilesBatchProcess::class, 'finishBatch'])
      ->setProgressMessage(\t('Completed @current of @total operations.'));
    foreach ($fileids as $file_id) {
      if ($file_id != 0) {
        $batch->addOperation([AuditFilesUsedNotManagedBatchProcess::class, 'auditfilesUsedNotManagedBatchDeleteProcessBatch'], [$file_id]);
      }
    }
    return $batch;
  }

  /**
   * Deletes the specified file from the file_usage table.
   *
   * @param int $file_id
   *   The ID of the file to delete from the database.
   */
  public function auditfilesUsedNotManagedBatchDeleteProcessFile($file_id): void {
    $num_rows = $this->connection->delete('file_usage')->condition('fid', $file_id)->execute();
    if (empty($num_rows)) {
      $this->messenger()->addWarning(
        t(
          'There was a problem deleting the record with file ID %fid from the file_usage table. Check the logs for more information.',
          ['%fid' => $file_id]
        )
      );
    }
    else {
      $this->messenger()->addStatus(
        t(
          'Sucessfully deleted File ID : %fid from the file_usages table.',
          ['%fid' => $file_id]
        )
      );
    }
  }

}
